from datetime import datetime, timezone
from pathlib import Path
from peewee import fn
import json
import pytest
from videobox import create_app
import videobox.sync as sync
from videobox.models import db_wrapper, Tag, Series, SeriesTag, Episode, Release
import videobox.models as models
from .conftest import TestingConfig

TEST_DIR = Path(__file__).parent

@pytest.fixture()
def app():
    app = create_app(data_dir=TEST_DIR, config_class=TestingConfig)
    with open(TEST_DIR.joinpath("test-data.sql"), 'r') as f:
        # https://docs.python.org/3/library/sqlite3.html#sqlite3.Connection.executescript
        sql = f.read()
        db_wrapper.database.connection().executescript(sql)

    yield app

    db_wrapper.database.connection().close()


def test_save_tags(app):
    with open(TEST_DIR.joinpath("sync", 'tags.json'), "r") as json_file:
        json_data = json.load(json_file)
        assert models.save_tags(app, json_data) == len(json_data)
        assert Tag.get_or_none(id=json_data[0]['id'])

def test_save_series(app):
    with open(TEST_DIR.joinpath("sync", 'series.json'), "r") as json_file:
        json_data = json.load(json_file)
        assert models.save_series(app, json_data) == len(json_data)
        assert Series.get_or_none(id=json_data[0]['id'])

def test_save_series_tags(app):
    with open(TEST_DIR.joinpath("sync", 'series-tags.json'), "r") as json_file:
        json_data = json.load(json_file)
        assert SeriesTag.get_or_none(SeriesTag.series==json_data[0]['series_id'], SeriesTag.tag==json_data[0]['tag_id'])

def test_save_episodes(app):
    with open(TEST_DIR.joinpath("sync", 'episodes.json'), "r") as json_file:
        json_data = json.load(json_file)
        assert models.save_episodes(app, json_data) == len(json_data)
        assert Episode.get_or_none(id=json_data[0]['id'])

def test_save_releases(app):
    with open(TEST_DIR.joinpath("sync", 'releases.json'), "r") as json_file:
        json_data = json.load(json_file)
        assert models.save_releases(app, json_data) == len(json_data)
        assert Release.get_or_none(id=json_data[0]['id'])



import webbrowser 
import os
from threading import Thread
import rumps
import waitress
import videobox
import videobox.sync as sync

APP_SERVER_PORT = 9157
APP_URL = f"http://127.0.0.1:{APP_SERVER_PORT}"

class VideoboxApp(rumps.App):
    def __init__(self, flask_app, name, title=None, icon=None, template=None, menu=None):
        self.flask_app = flask_app
        super().__init__(name, title, icon, template, menu)

    @rumps.clicked("Open Web Interface")
    def open_web_interface(self, _):
        webbrowser.open_new(APP_URL)

    @rumps.events.on_wake
    def on_wake(self):
        with self.flask_app.app_context():
            # Wait for the current worker to finish
            sync.sync_worker.abort()
            sync.sync_worker.join(videobox.MAX_WORKER_TIMEOUT)
            # Restart immediately with another worker
            sync.sync_worker = sync.SyncWorker(self.flask_app.config["API_CLIENT_ID"])
            sync.sync_worker.start()

    @rumps.events.before_quit
    def before_quit(self):
        videobox.shutdown_workers(self.flask_app)

if __name__ == "__main__":
    data_dir = rumps.application_support("Videobox")
    flask_app = videobox.create_app(app_dir=os.getcwd(), data_dir=data_dir)

    def runner():
        waitress.serve(flask_app, host='127.0.0.1', port=APP_SERVER_PORT, threads=8)

    # Start a local web server
    waitress_worker = Thread(name="Waitress worker", target=runner, daemon=False)
    waitress_worker.start()
    
    #VideoboxApp(flask_app, "Videobox", icon="menubar-icon.svg", template=True).run(debug=True)
    VideoboxApp(flask_app, "Videobox", icon="menubar-icon.svg", template=True).run()


from pathlib import Path
import pytest
from videobox import create_app

TEST_DIR = Path(__file__).parent

class TestingConfig(object):
    DATABASE_URL = 'sqlite:///:memory:'   
    API_CLIENT_ID = '123e4567-e89b-12d3-a456-426614174000'
    TESTING = True

@pytest.fixture()
def app():
    return create_app(config_class=TestingConfig)


from datetime import datetime, timedelta, timezone
from peewee import *
from playhouse.migrate import migrate, SqliteMigrator
from playhouse.reflection import Introspector
from playhouse.sqlite_ext import FTS5Model, SearchField, RowIDField
from playhouse.flask_utils import FlaskDB
from . import iso639

# @@TODO check 
# SQLite >3.32.0 has a limit of total 32766 max variables (SQLITE_MAX_VARIABLE_NUMBER)
# https://stackoverflow.com/a/64419474 
# https://stackoverflow.com/q/35616602
INSERT_CHUNK_SIZE = 999 // 15   # Series class has the max numbes of fields 

SYNC_STARTED = "S"
SYNC_ERROR = "E"
SYNC_OK = "K"

TAG_GENRE = "G"
TAG_KEYWORD = "K"

TRACKER_NOT_CONTACTED = 'N'
TRACKER_PROTOCOL_ERROR = 'P'
TRACKER_DNS_ERROR = 'D'
TRACKER_OK = 'K'
TRACKER_TIMED_OUT = 'T'

TRACKERS_ALIVE = [TRACKER_NOT_CONTACTED, TRACKER_OK, TRACKER_TIMED_OUT]

class AppDB(FlaskDB):
    '''
    Specialised FlaskDB which deals with testing memory 
      sqlite database, see: https://t.ly/susgy
    '''
    def _register_handlers(self, app):
        if app.config['TESTING']:
            return
        app.before_request(self.connect_db)
        app.teardown_request(self.close_db)
        
# Defer init in app creation
db_wrapper = AppDB()

class SyncLog(db_wrapper.Model):
    timestamp = TimestampField(utc=True)
    status = FixedCharField(max_length=1, default=SYNC_STARTED)
    description = TextField(default="")

    def __str__(self):
        return f"[{self.status} {self.timestamp}] {self.description}"


def get_last_log():
    return SyncLog.select().where(SyncLog.status != SYNC_STARTED).order_by(SyncLog.timestamp.desc()).get_or_none()


def get_last_successful_log():
    return SyncLog.select().where(SyncLog.status == SYNC_OK).order_by(SyncLog.timestamp.desc()).get_or_none()


class Series(db_wrapper.Model):
    tmdb_id = IntegerField(unique=True)
    imdb_id = CharField(default="")
    name = CharField()
    sort_name = CharField()
    tagline = CharField(default="")
    slug = CharField()
    overview = TextField()
    network = CharField(default="")
    poster_url = CharField(default="")
    fanart_url = CharField(default="")
    vote_average = FloatField(default=0)
    popularity = FloatField(default=0)
    status = FixedCharField(max_length=1)
    language = FixedCharField(max_length=2)
    followed_since = DateField(null=True)


    @property
    def poster(self):
        return self.poster_url or "/static/default-poster.png"
    

    @property
    def imdb_url(self):
        if self.imdb_id:
            return f"https://www.imdb.com/title/{self.imdb_id}/"
        else:
            return ""
        
    @property
    def tmdb_url(self):
        return f"https://www.themoviedb.org/tv/{self.tmdb_id}"

    def __str__(self):
        return self.name


class SeriesIndex(FTS5Model):
    """
    Full-text search index for series
    """
    rowid = RowIDField()
    name = SearchField()
    content = SearchField() # Contains all other relevant fields

    class Meta:
        database = db_wrapper.database
        options = {'tokenize': 'porter'}


class Tag(db_wrapper.Model):
    slug = CharField()
    name = CharField()
    type = FixedCharField(max_length=1)    

    def __str__(self):
        return self.name


class SeriesTag(db_wrapper.Model):
    series = ForeignKeyField(Series, on_delete="CASCADE")
    tag = ForeignKeyField(Tag, on_delete="CASCADE")

    class Meta:
        primary_key = CompositeKey('series', 'tag')


class Episode(db_wrapper.Model):
    tmdb_id = IntegerField()
    series = ForeignKeyField(Series, backref='episodes', on_delete="CASCADE")
    name = CharField()
    season = SmallIntegerField()
    number = SmallIntegerField()
    aired_on = DateField(null=True)
    overview = TextField(default="")
    thumbnail_url = CharField(default="")

    @property
    def season_episode_id(self):
        return "S{:02}.E{:02}".format(self.season, self.number)
    
    @property
    def thumbnail(self):
        return self.thumbnail_url or "/static/default-still.png"
    
    def __str__(self):
        return f"{self.season_episode_id} '{self.name}'"

    class Meta:
        indexes = (
            (('series', 'season', 'number'), True),
        )


# class EpisodeIndex(FTS5Model):
#     """
#     Full-text search index for episodes
#     """
#     rowid = RowIDField()
#     name = SearchField()
#     overview = SearchField()

#     class Meta:
#         database = db_wrapper.database
#         options = {'tokenize': 'porter'}

class Release(db_wrapper.Model):
    # Enough for BitTorrent 2 SHA-256 hashes
    info_hash = CharField(unique=True, max_length=64)
    # We could change an episode id, so update this FK accordingly
    episode = ForeignKeyField(
        Episode, backref='releases', on_update="CASCADE", on_delete="CASCADE")
    added_on = DateTimeField()
    last_updated_on = DateTimeField()
    size = BigIntegerField()
    magnet_uri = TextField()
    seeders = IntegerField()
    leechers = IntegerField()
    completed = IntegerField()
    name = CharField()
    resolution = SmallIntegerField()

    def __str__(self):
        return self.name

    # @property
    # def next_scrape_on(self):
    #     import videobox.scraper as scraper
    #     age = datetime.now(timezone.utc) - self.added_on.replace(tzinfo=timezone.utc)
    #     return scraper.get_scrape_threshold(age.days)

    @property
    def languages(self):
        return iso639.extract_languages(self.name)


class Tracker(db_wrapper.Model):
    url = CharField(unique=True)    
    status = FixedCharField(max_length=1, default=TRACKER_NOT_CONTACTED)
    last_scraped_on = DateTimeField(null=True)


def save_trackers(app, trackers):
    """
    Insert new trackers and ignore existing ones
    """    
    count = 0
    app.logger.debug("Saving trackers to database...")
    for batch in chunked(trackers, INSERT_CHUNK_SIZE):
        # Skip rows causing the conflict, see:
        #   https://sqlite.org/lang_conflict.html
        count += (Tracker.insert_many(batch)
                  .on_conflict_ignore()
                  .as_rowcount()
                  .execute())                
    return count

def save_tags(app, tags):
    """
    Insert new tags and attempt to update existing ones
    """
    count = 0
    app.logger.debug("Saving tags to database...")
    for batch in chunked(tags, INSERT_CHUNK_SIZE):
        count += (Tag.insert_many(batch)
                  # Replace current tag with new data
                  .on_conflict_replace()
                  .as_rowcount()
                  .execute())
    return count

def save_series(app, series):
    """
    Insert new series and attempt to update existing ones
    """
    count = 0
    app.logger.debug("Saving series to database...")
    for batch in chunked(series, INSERT_CHUNK_SIZE):
        count += (Series.insert_many(batch)
                  .on_conflict(
            conflict_target=[Series.id],
            # Pass down values from insert clause
            preserve=[Series.imdb_id, Series.name, Series.sort_name, Series.language, Series.tagline, Series.overview, Series.network,
                      Series.vote_average, Series.popularity, Series.poster_url, Series.fanart_url, Series.status])
                  .as_rowcount()
                  .execute())
        for series in batch:
            content = ' '.join([series['network'], series['overview']]) 
            # FTS5 insert_many cannot handle upserts
            (SeriesIndex.insert({
                SeriesIndex.rowid: series['id'],                    
                SeriesIndex.name: series['name'],
                SeriesIndex.content: content,
            })
                # Just replace name and content edits
                .on_conflict_replace()
                .execute())
    SeriesIndex.optimize()
    return count

def save_series_tags(app, series_tags):
    count = 0
    app.logger.debug("Saving series tags to database...")
    for batch in chunked(series_tags, INSERT_CHUNK_SIZE):
        count += (SeriesTag.insert_many(batch)
                  # Skip rows causing the conflict, see:
                  #   https://sqlite.org/lang_conflict.html
                  .on_conflict_ignore()
                  .execute())

    return count

def save_episodes(app, episodes):
    """
    Insert new episodes and attempt to update existing ones
    """
    count = 0
    app.logger.debug("Saving episodes to database...")
    for batch in chunked(episodes, INSERT_CHUNK_SIZE):
        # We need to cope with the unique constraint for (series, season, number)
        #   index because we cannot rely on episodes id's,
        #   they are often changed when TVDB users update them
        count += (Episode
                  .insert_many(batch)
                  .on_conflict(
                      conflict_target=[
                          Episode.series, Episode.season, Episode.number],
                      # Pass down values from insert clause
                      preserve=[Episode.name, Episode.overview,
                                Episode.aired_on, Episode.thumbnail_url])
                  .as_rowcount()
                  .execute())
        # EpisodeIndex.insert({
        #     EpisodeIndex.rowid: episode_id,
        #     EpisodeIndex.name: episode.name,
        #     EpisodeIndex.overview: episode.overview}).execute()
    #EpisodeIndex.optimize()            
    return count

def save_releases(app, releases):
    """
    Insert new releases and attempt to update existing ones
    """
    count = 0
    app.logger.debug("Saving releases to database...")
    for batch in chunked(releases, INSERT_CHUNK_SIZE):
        count += (Release
                  .insert_many(batch)
                  .on_conflict(
                      conflict_target=[Release.id],
                      # Pass down values from insert clause
                      preserve=[Release.leechers, Release.seeders, Release.completed, Release.last_updated_on])
                  .as_rowcount()
                  .execute())
    return count

###########
# DB SETUP
###########

def setup():
    db_wrapper.database.create_tables([
        Series,
        SeriesIndex,
        Episode,
        Release,
        Tracker,
        Tag,
        SeriesTag,
        SyncLog,
    ], safe=True)

    # Run schema update for every fields added after version 0.5

    migrator = SqliteMigrator(db_wrapper.database)
    introspector = Introspector.from_database(db_wrapper.database)
    models = introspector.generate_models()
    Series_ = models['series']
    Episode_ = models['episode']

    column_migrations = []
    
    # Add new columns

    if not hasattr(Series_, 'followed_since'):
        followed_since = DateField(null=True)
        column_migrations.append(migrator.add_column('series', 'followed_since', followed_since))

    # Remove obsolete columns

    if hasattr(Series_, 'last_updated_on'):
        column_migrations.append(migrator.drop_column('series', 'last_updated_on'))

    if hasattr(Episode_, 'last_updated_on'):
        column_migrations.append(migrator.drop_column('episode', 'last_updated_on'))

    # Run all migrations
    with db_wrapper.database.atomic():
        migrate(*column_migrations)

    return len(column_migrations)


'''
Custom Jinja Filters
'''
import operator
import itertools
from urllib.parse import urlparse
from datetime import datetime, timezone
from videobox import iso639

MIN_SEEDERS = 1

def human_date(value):
    return value.strftime("%b %d, %Y")

def timeline_date(value):
    return value.strftime("%a, %b %d")

def to_date(value):
    return datetime.strptime(value, '%Y-%m-%d')

def human_date_time(value):
    return value.strftime("%b %d, %Y at %H:%M")

def datetime_since(since_value, current_value):
    """
    Returns string representing "time since" e.g.
    3 days ago, 5 hours ago etc.
    """

    # Make sure we are comparing two TZ-aware values
    diff = current_value.replace(tzinfo=timezone.utc) - since_value.replace(tzinfo=timezone.utc)

    periods = (
        (diff.days // 365, "year", "years"),
        (diff.days // 30, "month", "months"),
        (diff.days // 7, "week", "weeks"),
        (diff.days, "day", "days"),
        (diff.seconds // 3600, "hour", "hours"),
        (diff.seconds // 60, "minute", "minutes"),
        (diff.seconds, "second", "seconds"),
    )

    for period, singular, plural in periods:
        if period:
            return "%d %s ago" % (period, singular if period == 1 else plural)

    return "just now"

# def timedelta(days):
#     periods = (
#         (days // 365, "year", "years"),
#         (days // 30, "month", "months"),
#         (days // 7, "week", "weeks"),
#         (days // 1, "day", "days"),
#     )

#     for period, singular, plural in periods:
#         if period:
#             return "in %d %s" % (period, singular if period == 1 else plural)

#     return "later today"
    
def islice(iterable, stop):
    return itertools.islice(iterable, stop)

def groupby_attrs(iterable, attr, *attrs):
    return itertools.groupby(iterable, key=operator.attrgetter(attr, *attrs))

def networks(value):
    pieces = value.split(", ")
    if len(pieces) > 1:
        return f"{pieces[0]} and others"
    else:
        return pieces[0]

def torrent_health(value):
    color = "hi"
    if MIN_SEEDERS <= value < 4:
        color = "low"
    elif 4 <= value < 7:
        color = "medium"
    return f'<span class="torrent-{color}">{value}</span>'

def lang(code):
    try:
        return iso639.LANGUAGES_SET_1[code]
    except KeyError:
        return ''

def pluralize(prefix, value):
    return f"{prefix}{'s' if value > 1 else ''}"


def nice_url(value): 
    pieces = urlparse(value)
    return pieces.netloc 

FILTERS = [
    human_date,
    timeline_date,
    human_date_time,
    torrent_health,
    networks,
    lang,
    islice,
    groupby_attrs,
    to_date,
    datetime_since,
    #timedelta,
    pluralize,
    nice_url,
]

def init_app(app):
    for func in FILTERS:
        app.add_template_filter(func)



# https://en.wikipedia.org/wiki/List_of_ISO_639_language_codes
import re

# Set 1 codes
LANGUAGES_SET_1 = {
    'aa': 'Afar',
    'ab': 'Abkhazian',
    'af': 'Afrikaans',
    'am': 'Amharic',
    'ar': 'Arabic',
    'as': 'Assamese',
    'ay': 'Aymara',
    'az': 'Azerbaijani',
    'ba': 'Bashkir',
    'be': 'Belarusian',
    'bg': 'Bulgarian',
    'bn': 'Bengali',
    'bo': 'Tibetan',
    'br': 'Breton',
    'bs': 'Bosnian',
    'ca': 'Catalan',
    'cs': 'Czech',
    'cy': 'Welsh',
    'da': 'Danish',
    'de': 'German',
    'dz': 'Dzongkha',
    'el': 'Greek',
    'en': 'English',
    'es': 'Spanish',
    'et': 'Estonian',
    'eu': 'Basque',
    'fa': 'Persian',
    'fi': 'Finnish',
    'fj': 'Fijian',
    'fo': 'Faroese',
    'fr': 'French',
    'ga': 'Irish',
    'gl': 'Galician',
    'gn': 'Guarani',
    'gu': 'Gujarati',
    'he': 'Hebrew',
    'hi': 'Hindi',
    'hr': 'Croatian',
    'hu': 'Hungarian',
    'hy': 'Armenian',
    'id': 'Indonesian',
    'is': 'Icelandic',
    'it': 'Italian',
    'ja': 'Japanese',
    'ka': 'Georgian',
    'kk': 'Kazakh',
    'km': 'Khmer',
    'kn': 'Kannada',
    'ko': 'Korean',
    'ku': 'Kurdish',
    'ky': 'Kyrgyz',
    'lb': 'Luxembourgish',
    'lo': 'Lao',
    'lt': 'Lithuanian',
    'lv': 'Latvian',
    'mg': 'Malagasy',
    'mk': 'Macedonian',
    'ml': 'Malayalam',
    'mn': 'Mongolian',
    'mr': 'Marathi',
    'ms': 'Malay',
    'mt': 'Maltese',
    'nb': 'Norwegian Bokmål',
    'ne': 'Nepali',
    'nl': 'Dutch',
    'nn': 'Norwegian Nynorsk',
    'no': 'Norwegian',
    'oc': 'Occitan',
    'or': 'Oriya',
    'pa': 'Punjabi',
    'pl': 'Polish',
    'ps': 'Pashto',
    'pt': 'Portuguese',
    'ro': 'Romanian',
    'ru': 'Russian',
    'rw': 'Kinyarwanda',
    'si': 'Sinhala',
    'sk': 'Slovak',
    'sl': 'Slovenian',
    'sq': 'Albanian',
    'sr': 'Serbian',
    'sv': 'Swedish',
    'sw': 'Swahili',
    'ta': 'Tamil',
    'te': 'Telugu',
    'th': 'Thai',
    'tl': 'Tagalog',
    'tr': 'Turkish',
    'uk': 'Ukrainian',
    'ur': 'Urdu',
    'vi': 'Vietnamese',
    'yi': 'Yiddish',
    'zh': 'Chinese',
    'zu': 'Zulu',
}

# Set 2 codes
LANGUAGES_SET_2 = {    
    "ita": "Italian",
    "fra": "French",
    "fre": "French",
    "spa": "Spanish",
    "deu": "German",
    "ger": "German",
    "eng": "English",
}

# Detect both three-characters and full English names
LANGUAGES_SET_X = LANGUAGES_SET_2 | {v.lower(): v for v in LANGUAGES_SET_2.values()}

RE_SEPARATOR = re.compile(r'[\.,:;\-_]')

def extract_languages(value):
    name = RE_SEPARATOR.sub(' ', value.lower())
    return [LANGUAGES_SET_X[piece] for piece in name.split() if (piece in LANGUAGES_SET_X)]



# See here for BitTorrent UDP scraping:  
#  https://www.bittorrent.org/beps/bep_0015.html
#  https://www.bittorrent.org/beps/bep_0041.html
# Adapted from https://github.com/erindru/m2t/blob/master/m2t/scraper.py

from operator import itemgetter
import binascii, socket, random, struct
from datetime import datetime, timedelta, timezone
import time
from urllib.parse import urlparse, parse_qs
#import urllib.error
from peewee import *
from flask import current_app as app
import videobox.models as models
from videobox.models import Series, Episode, Release, Tracker

UDP_TIMEOUT = 5
MAX_TORRENTS = 74               # UDP limit 
MAX_SEASONS = 2 
MIN_SCRAPING_INTERVAL = 1/24*3  # Days
MAX_SCRAPING_INTERVAL = 90      # Days
NEXT_RETRY_DAYS = 3      # Days

PROTOCOL_ID = 0x41727101980
ACTION_CONNECT = 0x0
ACTION_SCRAPE = 0x2
ACTION_ERROR = 0x3

class UnknownTrackerScheme(RuntimeError):
    pass

def scrape_tracker(tracker, hashes):
    """
    Returns the list of seeders, leechers and downloads a torrent info_hash has, according to the specified tracker
    """
    parsed = urlparse(tracker.lower())	
    if parsed.scheme == "udp":
        return scrape_udp(parsed, hashes)
    else:
        raise UnknownTrackerScheme(f"Unknown tracker scheme {parsed.scheme.upper()}, skipped")	

def scrape_udp(parsed_tracker, hashes):
    app.logger.debug(f"Scraping {parsed_tracker.geturl()} for {len(hashes)} hashes")
    if len(hashes) > MAX_TORRENTS:
        raise RuntimeError(f"Only {MAX_TORRENTS} hashes can be scraped on a UDP tracker")
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.settimeout(UDP_TIMEOUT)
    conn = (socket.gethostbyname(parsed_tracker.hostname), parsed_tracker.port)

    request, transaction_id = udp_create_connection_request()
    sock.sendto(request, conn)

    buf, _ = sock.recvfrom(2048)
    connection_id = udp_parse_connection_response(buf, transaction_id)

    # Scrape away
    request, transaction_id = udp_create_scrape_request(connection_id, hashes)
    sock.sendto(request, conn)
    buf, _ = sock.recvfrom(2048)
    data = udp_parse_scrape_response(parsed_tracker, buf, transaction_id, hashes)
    return data

def udp_create_connection_request():    	
    transaction_id = udp_get_transaction_id()
    buf = struct.pack("!q", PROTOCOL_ID) # First 8 bytes is a magic constant
    buf += struct.pack("!i", ACTION_CONNECT) # Next 4 bytes is action
    buf += struct.pack("!i", transaction_id) # Next 4 bytes is transaction id
    return (buf, transaction_id)

def udp_parse_connection_response(buf, sent_transaction_id):
    if len(buf) < 16:
        raise RuntimeError(f"Wrong response length getting connection id: {len(buf)}")			
    action = struct.unpack_from("!i", buf)[0] # First 4 bytes is action

    response_transaction_id = struct.unpack_from("!i", buf, 4)[0] # Next 4 bytes is transaction id
    if response_transaction_id != sent_transaction_id:
        raise RuntimeError(f"Transaction ID doesn't match in connection response: expected {sent_transaction_id}, got {response_transaction_id}")

    if action == ACTION_CONNECT:
        connection_id = struct.unpack_from("!q", buf, 8)[0] # Unpack 8 bytes from byte 8, should be the connection_id
        return connection_id
    elif action == ACTION_ERROR:		
        error = struct.unpack_from("!s", buf, 8)
        raise RuntimeError(f"Error {error} while trying to get a connection response")
    else:
        raise RuntimeError(f"Unknown action value: {action}")

def udp_create_scrape_request(connection_id, hashes):
    transaction_id = udp_get_transaction_id()
    buf = struct.pack("!q", connection_id) # First 8 bytes is connection id
    buf += struct.pack("!i", ACTION_SCRAPE) # Next 4 bytes is action 
    buf += struct.pack("!i", transaction_id) # Followed by 4 byte transaction id

    # From here on, there is a list of info_hashes. They are packed as char[]
    for hash in hashes:		
        hex_repr = binascii.a2b_hex(hash)
        buf += struct.pack("!20s", hex_repr)
    return (buf, transaction_id)

def udp_parse_scrape_response(parsed_tracker, buf, sent_transaction_id, hashes):	
    if len(buf) < 16:
        raise RuntimeError(f"Wrong response length while scraping: {len(buf)}")	

    action = struct.unpack_from("!i", buf)[0] # First 4 bytes is action
    response_transaction_id = struct.unpack_from("!i", buf, 4)[0] # Next 4 bytes is transaction id	
    if response_transaction_id != sent_transaction_id:
        raise RuntimeError(f"Transaction ID doesn't match in scrape response: expected {sent_transaction_id}, got {response_transaction_id}")
    if action == ACTION_SCRAPE:
        response = {}
        offset = 8
        # Next 4 bytes after action is transaction_id, so data doesn't start till byte 8		
        for hash in hashes:
            try: 
                seeds = struct.unpack_from("!i", buf, offset)[0]
                offset += 4
                completed = struct.unpack_from("!i", buf, offset)[0]
                offset += 4
                leeches = struct.unpack_from("!i", buf, offset)[0]
                offset += 4			
            except struct.error:
                raise RuntimeError("Error while unpacking scrape response from tracker")
            response[hash] = {
                "seeders": seeds, 
                "leechers": leeches, 
                "completed": completed, 
                'tracker': parsed_tracker.hostname
            }
        return response
    elif action == ACTION_ERROR:
        # Error occured, try and extract the error string
        error = struct.unpack_from("!s", buf, 8)
        raise RuntimeError(f"Error while scraping: {error}")
    else:
        raise RuntimeError(f"Unknown action value: {action}")

def udp_get_transaction_id():
    return random.randint(0, 255)

def make_series_subquery():
    SeriesAlias = Series.alias()
    return (SeriesAlias.select(SeriesAlias, fn.Max(Episode.season).alias("max_season"))
            .join(Episode)
            .group_by(SeriesAlias.id))

def get_releases():
    since_datetime = datetime.now(timezone.utc) - timedelta(days=MAX_SCRAPING_INTERVAL)
    series_subquery = make_series_subquery()
    return (Release.select(Release)
            .join(Episode)
            .join(Series)
            .join(series_subquery, on=(
                series_subquery.c.id == Series.id))
            # Do not bother to scrape releases from old seasons
            .where((series_subquery.c.max_season-Episode.season < MAX_SEASONS) & 
                   (Release.added_on >= since_datetime) & 
                   # Sqlite 'now' uses UTC, see https://www.sqlite.org/lang_datefunc.html
                   (fn.JulianDay('now') - fn.JulianDay(Release.last_updated_on) >
                    (fn.Threshold(fn.JulianDay('now') - fn.JulianDay(Release.added_on)))))
            # Scrape recent releases first
            .order_by(Release.added_on.desc()))

def scrape_releases(): 
    start = time.time()
    releases = get_releases()
    trackers = collect_trackers(releases)
    models.save_trackers(app, [{'url': tracker} for tracker in trackers])
    # Contact less frequently those trackers which return fatal errors
    next_attempt_days = Case(Tracker.status, (
        (models.TRACKER_PROTOCOL_ERROR, NEXT_RETRY_DAYS),
        (models.TRACKER_DNS_ERROR, NEXT_RETRY_DAYS)), 0)
    candidate_trackers = [tracker.url for tracker in Tracker.select().where((Tracker.last_scraped_on == None) | (fn.JulianDay('now') - fn.JulianDay(Tracker.last_scraped_on) >= next_attempt_days))]
    app.logger.info(f"Start scraping {len(candidate_trackers)} trackers...")
    # Remove TZ or Peewee will save it as string in SQLite
    utc_now = datetime.now(timezone.utc).replace(tzinfo=None)    
    scraped_torrents = {}
    for tracker_url, info_hashes in trackers.items():
        if tracker_url not in candidate_trackers:
            continue
        status = models.TRACKER_PROTOCOL_ERROR
        try:
            torrents = {}
            for chunked_info_hashes in chunked(info_hashes, MAX_TORRENTS):
                torrents.update(scrape_tracker(tracker_url, chunked_info_hashes))
                # Do not flood server with requests
                time.sleep(0.75)
            status = models.TRACKER_OK
        except UnknownTrackerScheme as ex:
            app.logger.debug(ex)
            continue
        except socket.gaierror:
            app.logger.debug(f"{tracker_url} name or service is not know, skipped")
            status = models.TRACKER_DNS_ERROR
            continue
        except ConnectionRefusedError:
            app.logger.debug(f"Connection to {tracker_url} was refused, skipped")
            continue
        except RuntimeError as ex:
            app.logger.debug(f"Request to {tracker_url} gave an exception ({ex}), skipped")
            continue
        except socket.timeout:
            app.logger.debug(f"Request to {tracker_url} timed out, skipped")
            status = models.TRACKER_TIMED_OUT
            # Do not skip, but collect as much scraped data as possible
        finally:
            Tracker.update(last_scraped_on=utc_now, status=status).where(Tracker.url == tracker_url).execute()

        # Group scraped data by info_hash
        for info_hash, data in torrents.items():
            scraped_torrents.setdefault(info_hash, []).append(data)

    for info_hash, all_data in scraped_torrents.items():
        # Get best seeds result for each torrent
        data = max(all_data, key=itemgetter("seeders"))
        Release.update(seeders=data['seeders'], 
                       leechers=data['leechers'], 
                       completed=data['completed'],
                       last_updated_on=utc_now).where(Release.info_hash == info_hash).execute()
    end = time.time()
    app.logger.info(f"Scraped {len(scraped_torrents)} of {len(releases)} releases in {end-start:.1f}s.")


def get_magnet_uri_trackers(magnet_uri):
    pieces = urlparse(magnet_uri)
    if pieces.scheme != 'magnet':
        raise ValueError(f"Invalid valid magnet link {magnet_uri}")
    data = parse_qs(pieces.query)
    return map(str.lower, data['tr'])

def get_scrape_threshold(value):
    # Normalise (0, max) range to (0, 1) and use a quad function,
    #   see https://easings.net/#easeInQuad
    value = 1 - (MAX_SCRAPING_INTERVAL-value) / MAX_SCRAPING_INTERVAL
    return max(MIN_SCRAPING_INTERVAL, value*value*MAX_SCRAPING_INTERVAL)

def collect_trackers(releases):
    trackers = {}
    for release in releases:
        try:
            tracker_urls = get_magnet_uri_trackers(release.magnet_uri)
        except ValueError as ex:
            app.logger.debug(f"{ex}, skipped")
            continue            
        # Group all torrents by their tracker URL's, 
        #   to minimize newtwork calls
        for url in tracker_urls:
            trackers.setdefault(url, []).append(release.info_hash)
    return trackers


from datetime import datetime, date, timezone
from operator import attrgetter
from itertools import groupby
import re
import flask
# from flask import current_app as app
from peewee import fn, JOIN
from playhouse.flask_utils import PaginatedQuery, get_object_or_404
import videobox
import videobox.models as models
from videobox.models import Series, Episode, Release, Tag, SeriesTag, SyncLog, Tracker
from . import bp
from .announcer import announcer
from . import queries

MAX_CHART_DAYS = 30 
MAX_TOP_TAGS = 8
MIN_SEEDERS = 1
MAX_SEASONS = 2
MAX_LOG_ROWS = 3
SERIES_CARDS_PER_PAGE = 6 * 10
SERIES_EPISODES_PER_PAGE = 30
RESOLUTION_OPTIONS = {
    0: "Any",
    480: "480p",
    720: "720p (HD)",
    1080: "1080p (Full HD)",
    2160: "2160p (4K)",
}
SIZE_OPTIONS = {
    "": "Any",
    "asc": "Smallest",
    "desc": "Largest",
}
RE_INFO_HASH = re.compile(r"^[0-9a-fA-F]{40}$")


@bp.context_processor
def inject_template_vars():
    return {
        "version": videobox.__version__
    }


@bp.errorhandler(404)
def page_not_found(e):
    return flask.render_template('404.html'), 404


# ---------
# Home
# ---------

last_server_alert = ''

@bp.route('/')
def home():
    total_series, total_episodes, total_releases = queries.get_library_stats()
    # Make sure library is already filled with data
    if total_series and total_episodes and total_releases:        
        last_sync = models.get_last_log()
        utc_now = datetime.now(timezone.utc)
        today_series = queries.get_today_series(10)
        # Do not exlude any series for now
        exclude_ids = []
        featured_series = queries.get_featured_series(exclude_ids=exclude_ids, days_interval=2).limit(8)
        top_tags = queries.get_top_tags(MAX_TOP_TAGS)
        # Show updates within the last week
        followed_series = queries.get_followed_series(days=7)
        return flask.render_template("home.html", 
                                     last_sync=last_sync,
                                     utc_now=utc_now,
                                     server_alert=last_server_alert,
                                     today_series=today_series,
                                     featured_series=featured_series, 
                                     top_tags=top_tags,
                                     total_series=total_series,
                                     total_releases=total_releases,
                                     followed_series=followed_series)
    else:
        return flask.render_template("first-import.html")

# ---------
# Search
# ---------


@bp.route('/search')
def search():
    query = flask.request.args.get("query")
    if not query:
        flask.abort(400)
    if is_info_hash(query):
        release = get_object_or_404(Release, (Release.info_hash==query.lower()))
        return flask.redirect(flask.url_for('.series_detail', series_id=release.episode.series.id, view="list", _anchor=f"r{release.info_hash}"))
    else:
        series_ids = [series.rowid for series in queries.search_series(
            sanitize_query(query))]
        series = queries.get_series_with_ids(series_ids)
        if len(series) == 1:
            return flask.redirect(flask.url_for('.series_detail', series_id=series[0].id))        
        else:
            return flask.render_template("search_results.html", found_series=series, search_query=query)


def is_info_hash(value):
    return RE_INFO_HASH.match(value)


@bp.route('/suggest')
def suggest():
    query = flask.request.args.get("query")
    if not query:
        flask.abort(400)
    sanitized_query = sanitize_query(query)
    search_suggestions = queries.suggest_series(sanitized_query).limit(10)
    return flask.render_template("_suggest.html", search_suggestions=search_suggestions)


# ---------
# Tag
# ---------

@bp.route('/tag')
def tag():    
    tags_all = queries.get_top_series_for_tags()
    tags_series = groupby(tags_all, key=attrgetter('tag_slug', 'tag_name'))
    return flask.render_template("tags.html", tags_series=tags_series)


@bp.route('/tag/<slug>')
def tag_detail(slug):
    page = flask.request.args.get("page", 1, type=int)
    tag = get_object_or_404(Tag, (Tag.slug == slug))
    query = queries.get_series_for_tag(tag)
    paginated_series = PaginatedQuery(query, paginate_by=SERIES_CARDS_PER_PAGE, page_var="page", check_bounds=True)
    if page == 1:
        return flask.render_template("tag_detail.html", tag=tag, series=paginated_series, page=page, series_count=query.count())
    else:
        # For async requests
        return flask.render_template("_tag-card-grid.html", tag=tag, series=paginated_series, page=page)

# ---------
# Languages
# ---------

@bp.route('/language/<code>')
def language_detail(code):
    page = flask.request.args.get("page", 1, type=int)
    query = queries.get_series_for_language(code)
    paginated_series = PaginatedQuery(query, paginate_by=SERIES_CARDS_PER_PAGE, page_var="page", check_bounds=True)
    if page == 1:
        return flask.render_template("language_detail.html", language=code, series=paginated_series, page=page, series_count=query.count())
    else:
        # For async requests
        return flask.render_template("_language-card-grid.html", language=code, series=paginated_series, page=page)


# --------- 
# Series Detail
# ---------


@bp.route('/series/<int:series_id>')
def series_detail(series_id):
    series = get_object_or_404(Series, (Series.id == series_id))
    return _series_detail(series)


def _series_detail(series):
    #resolution = flask.request.args.get("resolution", type=int, default=0) or flask.request.cookies.get('resolution', type=int, default=0)
    resolution = flask.request.args.get("resolution", type=int, default=0)
    #size_sorting = flask.request.args.get("size", default="") or flask.request.cookies.get('size', default="")
    size_sorting = flask.request.args.get("size", default="")
    episode_sorting = flask.request.args.get("episode", default="asc")
    view_layout = flask.request.args.get("view", default="grid")
    today = date.today()
    series_subquery = queries.get_series_subquery()
    release_cte = queries.release_cte(resolution, size_sorting)
    if resolution or size_sorting:
        episodes_query = (Episode.select(Episode, Release.id, Release.info_hash, Release.name, Release.magnet_uri, Release.resolution, Release.size, Release.seeders, Release.last_updated_on)
                          .join(Release)
                          .switch(Episode)
                          .join(Series)
                          .join(series_subquery, on=(
                              series_subquery.c.id == Series.id))
                          .join(release_cte, on=(Release.id == release_cte.c.release_id))
                          .where((Episode.series == series.id) &
                                 #(Release.seeders >= MIN_SEEDERS) &
                                 # Episodes from last 2 seasons only
                                 (series_subquery.c.max_season - Episode.season < MAX_SEASONS) 
                                 #& (Episode.aired_on != None)
                                 )
                          .order_by(Episode.season.desc(), Episode.number if episode_sorting == "asc" else Episode.number.desc())
                          .with_cte(release_cte))
    else:
        # Unfiltered
        episodes_query = (Episode.select(Episode, Release.id, Release.info_hash, Release.name, Release.magnet_uri, Release.resolution, Release.size, Release.seeders, Release.last_updated_on)
                          .join(Release, JOIN.LEFT_OUTER)
                          .switch(Episode)
                          .join(Series)
                          .join(series_subquery, on=(
                              series_subquery.c.id == Series.id))
                          .where((Episode.series == series.id) &
                                 #(Release.seeders >= MIN_SEEDERS) &
                                 # Episodes from last 2 seasons only
                                 (series_subquery.c.max_season - Episode.season < MAX_SEASONS) 
                                 # & (Episode.aired_on != None)
                                 )
                          .order_by(Episode.season.desc(), Episode.number if episode_sorting == "asc" else Episode.number.desc(), Release.seeders.desc()))

    # Group by season number
    seasons_episodes = groupby(episodes_query, key=attrgetter('season'))
    series_tags = queries.get_series_tags(series) 

    response = flask.make_response(flask.render_template("series_detail.html", 
                                                         series=series, 
                                                         series_tags=series_tags, 
                                                         seasons_episodes=seasons_episodes, 
                                                         today=today, 
                                                         resolution=resolution, 
                                                         resolution_options=RESOLUTION_OPTIONS, 
                                                         size=size_sorting, 
                                                         size_options=SIZE_OPTIONS,
                                                         episode_sorting=episode_sorting,
                                                         view_layout=view_layout))
    # Remember filters across requests
    if resolution:
        response.set_cookie('resolution', str(resolution))
    if size_sorting:
        response.set_cookie('size', size_sorting)    
    return response


@bp.route('/series/follow/<int:series_id>', methods=['POST'])
def series_detail_update(series_id):
    following = flask.request.form.get("following", type=int)
    series = get_object_or_404(Series, (Series.id == series_id))
    series.followed_since = date.today() if following else None
    series.save()

    # Toggle button
    return flask.render_template("_follow-button.html", series=series)

@bp.route('/release/<int:release_id>')
def release_detail(release_id):
    release = get_object_or_404(Release, (Release.id == release_id))
    return flask.render_template("_release_detail.html", utc_now=datetime.now(timezone.utc), release=release)


@bp.route('/following')
def following():
    page = flask.request.args.get("page", 1, type=int)
    query = queries.get_followed_series()
    paginated_series = PaginatedQuery(query, paginate_by=SERIES_EPISODES_PER_PAGE, page_var="page")
    if page == 1:
        return flask.render_template("following.html", paginated_series=paginated_series, page=page)
    else:
        # For async requests
        return flask.render_template("_following.html", paginated_series=paginated_series, page=page)
    
# ---------
# Sync database
# ---------

@bp.route('/sync/events')
def sync_events():    
    def stream():
        # Return a message queue
        messages = announcer.listen()
        while True:
            # Blocks until a new message arrives
            msg = messages.get()
            if announcer.is_close_message(msg):
                break
            yield msg
    return flask.Response(stream(), mimetype='text/event-stream')

# ---------
# System status
# ---------

@bp.route('/status')
def system_status():
    log_rows = SyncLog.select().order_by(SyncLog.timestamp.desc()).limit(MAX_LOG_ROWS)
    chart_query = Release.raw(f'SELECT DATE(added_on) AS release_date, COUNT(id) AS release_count FROM `release` GROUP BY release_date ORDER BY release_date DESC LIMIT {MAX_CHART_DAYS}')
    # Filter out trackers that will likely never reply correctly
    trackers = Tracker.select().where((Tracker.status << models.TRACKERS_ALIVE)).order_by(Tracker.status, Tracker.url)
    max_last_scraped_on = Tracker.select(fn.Max(Tracker.last_scraped_on).alias("max_last_scraped_on")).where((Tracker.status << [models.TRACKER_OK, models.TRACKER_TIMED_OUT])).scalar()
    return flask.render_template("status.html", 
                                 log_rows=log_rows, trackers=trackers, chart=chart_query, max_chart_days=MAX_CHART_DAYS, max_log_rows=MAX_LOG_ROWS, max_last_scraped_on=max_last_scraped_on)


# ---------
# Helpers
# ---------

def sanitize_query(query):
    # https://www.sqlite.org/fts5.html
    sanitized_query = ""
    for c in query:
        if c.isalpha() or c.isdigit() or ord(c) > 127 or ord(c) == 96 or ord(c) == 26:
            # Allowed in FTS queries
            sanitized_query += c
        else:
            sanitized_query += " "
    return sanitized_query


import pytest
from videobox import iso639

@pytest.mark.parametrize("value,expected", [
    ("Constellation.S01E01.L.angelo.ferito.ITA.ENG.1080p.ATVP.WEB.DL.DD5.1.H.264.MeM.GP.mkv", ["Italian", "English"]),
    ("Halo.S02e05..1080p.Ita.Eng.Spa.h265.10bit.SubS..byMe7alh..MIRCrew.", ["Italian", "English", "Spanish"]), 
    ("Home.Economics.S02E20.720p.HDTV.x264-SYNCOPY[eztv.re].mkv", []), 
    ("Star Wars: The Bad Batch S03E07 FRENCH HDTV", ["French"]), 
])
def test_extract_languages(app, value, expected):
    assert iso639.extract_languages(value) == expected



import time
from datetime import datetime, timedelta, timezone
from threading import Thread, Event
from peewee import chunked, fn
from requests.exceptions import HTTPError, ReadTimeout
from flask import current_app
import videobox.api as api
import videobox.models as models
import videobox.scraper as scraper
from videobox.models import SyncLog

REQUEST_CHUNK_SIZE = 450        # Total URI must be < 4096
TIMEOUT_BEFORE_RETRY = 5        # Seconds
SYNC_INTERVAL = 60*60*2         # Seconds
MIN_SYNC_INTERVAL = 60*15       # Seconds

# The only sync worker tread
sync_worker = None


class SyncError(Exception):
    pass

def default_progress_callback(message):
    pass

def default_done_callback(message, alert):
    pass

class SyncWorker(Thread):

    def __init__(self, client_id, progress_callback=default_progress_callback, done_callback=default_done_callback):
        super().__init__(name="Sync worker")
        self.app = current_app._get_current_object()
        self.client_id = client_id
        self.progress_callback = progress_callback
        self.done_callback = done_callback
        # Start sync immediately at startup
        self.interval = 0
        self.abort_event = Event()        

    def abort(self):
        self.abort_event.set()

    def run(self):
        # Set up a recurring execution unless asked to abort
        while not self.abort_event.is_set():
            if self.interval:
                self.app.logger.debug(f"Waiting for {self.interval}s before next sync...")
            self.abort_event.wait(self.interval)
            if self.abort_event.is_set():
                self.app.logger.debug(f"Stopped {self.name} #{id(self)} thread")
                return             
            self._run_sync()            
            # Schedule next sync
            self.interval = SYNC_INTERVAL

    def _run_sync(self):
        last_log = models.get_last_successful_log()
        start_time = time.time()

        # Manually push the app context to make Flask
        #   logger to work on the separate thread
        with self.app.app_context():
            if last_log:
                sync_interval = datetime.now(timezone.utc) - last_log.timestamp.replace(tzinfo=timezone.utc) 
                # timedelta.seconds upper bound is 3600*24
                if sync_interval.days == 0 and sync_interval.seconds < MIN_SYNC_INTERVAL:
                    self.app.logger.info(f"Sync request is below min. time interval of {MIN_SYNC_INTERVAL}s, ignored")
                    return

            current_log = SyncLog.create(description="Started sync")

            alert = ""
            try:
                if last_log:
                    alert, tags_count, series_count, episode_count, release_count = self.update_library(
                        last_log)
                else:
                    tags_count, series_count, episode_count, release_count = self.import_library()
            except SyncError as ex:
                self.update_log(current_log, status=models.SYNC_ERROR, description=str(ex))
                self.done_callback(str(ex), alert)
                return

            elapsed_time = time.time()-start_time
            if any([series_count, episode_count, release_count]):
                description = f"Added/updated {tags_count} tags, {series_count} series, {episode_count} episodes, and {release_count} torrents"
            else:
                description = "No updates were found"

            # Mark import/sync successful
            self.update_log(current_log, status=models.SYNC_OK, description=description)

            self.app.logger.info(f"Finished in {elapsed_time:.1f}s: {description}")

            self.done_callback(description, alert)

            scraper.scrape_releases()

    def import_library(self):
        tags_count, series_count, episode_count, release_count = 0, 0, 0, 0

        self.app.logger.info("No local database found, starting full import")

        self.progress_callback("Importing all tags...")

        json = self.do_json_request(
            lambda: api.get_all_tags(self.client_id), retries=3)
        if json:
            self.progress_callback("Saving tags to library...")
            tags_count = models.save_tags(self.app, json)

        self.progress_callback("Importing all series...")

        json = self.do_json_request(
            lambda: api.get_all_series(self.client_id))
        if json:
            self.progress_callback("Saving series to library...")
            series_count = models.save_series(self.app, json)

        self.progress_callback("Importing all series tags...")

        json = self.do_json_request(
            lambda: api.get_all_series_tags(self.client_id))
        if json:
            models.save_series_tags(self.app, json)

        self.progress_callback("Importing all episodes...")

        json = self.do_json_request(
            lambda: api.get_all_episodes(self.client_id))
        if json:
            self.progress_callback("Saving episodes to library...")
            episode_count = models.save_episodes(self.app, json)

        self.progress_callback("Importing all torrents...")

        json = self.do_json_request(
            lambda: api.get_all_releases(self.client_id))
        if json:
            self.progress_callback("Saving torrents to library...")            
            release_count = models.save_releases(self.app, json)

        return tags_count, series_count, episode_count, release_count

    def update_library(self, last_log):
        tags_count, series_count, episode_count, release_count = 0, 0, 0, 0

        self.app.logger.info("Last update done at {0} UTC, requesting updates since then".format(
            last_log.timestamp.isoformat()))
        self.progress_callback("Getting updated series...")
        # Ensure UTC timezone
        json = self.do_json_request(lambda: api.get_updated_series(
            self.client_id, last_log.timestamp.replace(tzinfo=timezone.utc)), retries=3)

        # Save alert from server, if any
        alert = json["alert"]

        tag_ids = json['tags']
        if tag_ids:
            self.app.logger.debug(
                "Got {0} tags, starting update".format(len(tag_ids)))
            tags_count = self.sync_tags(tag_ids)        

        # Grab series
        series_ids = json['series']
        if series_ids:
            self.app.logger.debug(
                "Got {0} series, starting update".format(len(series_ids)))
            series_count = self.sync_series(series_ids)

        # Grab episodes
        episode_ids = json['episodes']
        if episode_ids:
            self.app.logger.debug(
                "Got {0} episodes, starting update".format(len(episode_ids)))
            episode_count = self.sync_episodes(episode_ids)

        # Grab releases
        release_ids = json['releases']
        if release_ids:
            self.app.logger.debug(
                "Got {0} releases, starting update".format(len(release_ids)))
            release_count = self.sync_releases(release_ids)

        return alert, tags_count, series_count, episode_count, release_count

    def sync_tags(self, remote_ids):
        count = 0

        # Always request all remote ids so we have a chance to update existing series
        if remote_ids:
            def callback(remaining):
                self.progress_callback(
                    f"Updating {remaining} tags...")

            # Request all remote tags
            response = self.do_chunked_request(
                api.get_tags_with_ids, remote_ids, callback)
            if response:
                count = models.save_tags(self.app, response)

        return count

    def sync_series(self, remote_ids):

        # @@TODO
        # local_count = (Series.select(fn.Count(Series.id))
        #                .where((Series.id << remote_ids))
        #                .scalar())
        # missing_count = len(remote_ids) - local_count
        count, missing_count = 0, 0

        # Always request all remote ids so we have a chance to update existing series
        if remote_ids:
            def callback(remaining):
                self.progress_callback(
                    f"Updating {remaining} series...")

            # self.app.logger.debug(
            #     f"Found missing {missing_count} of {len(remote_ids)} series")
            # Request old and new series
            response = self.do_chunked_request(
                api.get_series_with_ids, remote_ids, callback)
            if response:
                count = models.save_series(self.app, response)

            #  Series tags
            response = self.do_chunked_request(
                api.get_series_tags_for_ids, remote_ids, callback)
            if response:
                models.save_series_tags(self.app, response)

        return count



    def sync_episodes(self, remote_ids):

        # local_ids = [e.id for e in Episode.select(Episode.id)]
        # new_ids = set(remote_ids) - set(local_ids)
        count, missing_count = 0, 0

        # Always request all remote ids so we have a chance to update existing episodes
        if remote_ids:
            def callback(remaining):
                self.progress_callback(
                    f"Updating {remaining} episodes...")

            # self.app.logger.debug(
            #     f"Found missing {missing_count} of {len(remote_ids)} episodes")
            # Request old and new episodes
            response = self.do_chunked_request(
                api.get_episodes_with_ids, remote_ids, callback)
            if response:
                count = models.save_episodes(self.app, response)

        return count

    def sync_releases(self, remote_ids):

        # local_ids = [r.id for r in Release.select(Release.id)]
        # new_ids = set(remote_ids) - set(local_ids)
        count, missing_count = 0, 0

        # Always request all remote ids so we have a chance to update existing releases
        if remote_ids:
            def callback(remaining):
                self.progress_callback(
                    f"Updating {remaining} torrents...")

            # self.app.logger.debug(
            #     f"Found missing {missing_count} of {len(remote_ids)} releases")
            # Request old and new releases
            response = self.do_chunked_request(
                api.get_releases_with_ids, remote_ids, callback)
            if response:
                count = models.save_releases(self.app, response)

        return count

    def update_log(self, log, status, description):
        log.status = status
        log.description = description
        log.save()

    def do_chunked_request(self, handler, ids, callback=None):
        result = []
        ids_count = len(ids)
        for index, chunked_ids in enumerate(chunked(ids, REQUEST_CHUNK_SIZE)):
            if callback:
                callback(ids_count - index*REQUEST_CHUNK_SIZE)
            self.app.logger.debug(
                f"Requesting {index + 1} of {ids_count // REQUEST_CHUNK_SIZE + 1} chunks")
            json = self.do_json_request(
                lambda: handler(self.client_id, chunked_ids))
            result.extend(json)
        return result

    def do_json_request(self, handler, retries=1):
        for index in reversed(range(retries)):
            try:
                response = handler()
                response.raise_for_status()  # Raise an exeption on HTTP errors
            except ReadTimeout as ex:
                message = f'Server timed out while handling the request, {"retrying" if index else "skipped"}'
                self.app.logger.warn(message)
                time.sleep(TIMEOUT_BEFORE_RETRY)
                continue  # Next retry
            except HTTPError as ex:
                message = f'Server error {ex.response.status_code} occurred while handling the request, giving up'
                self.app.logger.error(message)
                raise SyncError(message)
            return response.json()
        # No more retries, giving up
        raise SyncError(
            "Server timed out while handling the request")


from setuptools import setup
import videobox

APP = ['macos/app.py']
DATA_FILES = ['macos/menubar-icon.svg']
OPTIONS = {
    # List here specific app packages, in addition tp those in pyproject.toml
    'packages': ['rumps', 'videobox'],
    # Copy these files outside the app bundle lib/python39.zip archive
    'resources': ['./videobox/templates', './videobox/static'], 
    # Tweak app Info.plist file
    'plist': {
        'CFBundleIdentifier': 'com.passiomatic.videobox',
        'LSUIElement': True,
        'CFBundleShortVersionString': videobox.__version__,
        'NSHumanReadableCopyright': '© Andrea Peltrin'
    },
    'iconfile': 'build/icon.icns'
}

setup(
    name='Videobox',
    app=APP,
    data_files=DATA_FILES,
    options={'py2app': OPTIONS},
    setup_requires=['py2app'],
)



import queue

close_message = object()

class MessageAnnouncer(object):
    """
    Server-sent events in Flask without extra dependencies.
    See https://maxhalford.github.io/blog/flask-sse-no-deps/
    """

    def __init__(self):
        self.listeners = []

    def listen(self):
        q = queue.Queue(maxsize=5)
        self.listeners.append(q)
        return q

    def format_sse(self, data, event=None):
        msg = f'data: {data}\n\n'
        if event:
            msg = f'event: {event}\n{msg}'
        return msg

    def close(self):
        self.announce(close_message)

    def is_close_message(self, msg):
        return msg is close_message

    def announce(self, msg):
        for i in reversed(range(len(self.listeners))):
            try:
                self.listeners[i].put_nowait(msg)
            except queue.Full:
                del self.listeners[i]

announcer = MessageAnnouncer()


import operator
from functools import reduce
from datetime import datetime, date, timedelta, timezone
from peewee import fn
from videobox.models import Series, Episode, Release, Tag, SeriesTag, SeriesIndex, TAG_GENRE

MAX_SEASONS = 2

def get_series_tags(series):
    return (Tag
            .select()
            .join(SeriesTag)
            .join(Series)
            .where((Series.id == series.id) & (Tag.type == TAG_GENRE)))    

def get_library_stats():
    # Make sure we grab series and episodes with at least one release
    return (Series.select(fn.Count(Series.id.distinct()), fn.Count(Episode.id.distinct()), fn.Count(Release.id.distinct()))
            .join(Episode)
            .join(Release).scalar(as_tuple=True))

def get_top_tags(limit):
    return (Episode.select(Episode, Tag.name.alias('tag_name'), Tag.slug.alias("tag_slug"))
            .join(Series)
            .join(SeriesTag)
            .join(Tag)
            .where((Tag.type == TAG_GENRE) & (Episode.thumbnail_url != ''))
            .group_by(Tag.slug)
            .order_by(fn.Count(Series.id.distinct()).desc())
            # Do not show empty tags
            .having(fn.Count(Series.id.distinct()) > 0)
            .limit(limit)
            .objects()
            )

def get_series_subquery():
    SeriesAlias = Series.alias()
    return (SeriesAlias.select(SeriesAlias.id, fn.Max(Episode.season).alias("max_season"))
            .join(Episode)
            .group_by(SeriesAlias.id))


def get_featured_series(exclude_ids, days_interval):
    series_subquery = get_series_subquery()
    return (Series.select(Series, fn.Sum(Release.completed).alias('total_completed'))
            .join(Episode)
            .join(Release)
            .join(series_subquery, on=(
                series_subquery.c.id == Series.id))
            # Consider episodes from last season only and releases within interval days
            .where(~(Series.id << exclude_ids) & 
                   (Episode.season == series_subquery.c.max_season) &
                   (Release.added_on > (date.today() - timedelta(days=days_interval))))
            .group_by(Series)
            .order_by(fn.Sum(Release.completed).desc())
            )


def get_today_series(limit):
    series_subquery = get_series_subquery()
    return (Series.select(Series, Episode, fn.SUM(Release.completed).alias('total_completed'))
            .join(Episode)
            .join(Release)
            .join(series_subquery, on=(
                series_subquery.c.id == Series.id))
            # Consider episodes from last season only and releases within the past 24h
            .where((Episode.season == series_subquery.c.max_season) &
                   (Episode.thumbnail_url != '') &
                   # @@TODO do not user current time, figure out max added_on and compute from it
                   (Release.added_on > (datetime.now(timezone.utc) - timedelta(hours=24))))
            .order_by(fn.Sum(Release.completed).desc())
            .group_by(Series.id)
            #.get_or_none()
            .limit(limit)
            )

def get_followed_series(days=None):
    series_subquery = get_series_subquery()
    where_clauses = [
        # Only episodes from the last season        
        (Episode.season == series_subquery.c.max_season),
        (fn.strftime('%Y-%m-%d', Release.added_on) >= Series.followed_since)
    ]
    if days:
        min_date = date.today() - timedelta(days=days)
        where_clauses.append((fn.strftime('%Y-%m-%d', Release.added_on) > min_date))
    q = (Series.select(Series, Episode, fn.strftime('%Y-%m-%d', Release.added_on).alias("added_on_date"), fn.Count(Release.id).alias("release_count"))
         .join(Episode)
         .join(Release)
         .join(series_subquery, on=(
             series_subquery.c.id == Series.id))
         # Chain all conditions together AND'ing them 
         #  https://github.com/coleifer/peewee/issues/391#issuecomment-468042229
         .where(reduce(operator.and_, where_clauses))
         .group_by(Episode.id)
         # Make sure series' releases are one after another when listing
         .order_by(fn.strftime('%Y-%m-%d', Release.added_on).desc(), Series.id)
         )    
    return q

def get_top_series_for_tags():
    # Grab all tags and associated series with releases for the last two seasons
    subquery = get_series_subquery()
    return (Series.select(Series, Tag.slug.alias('tag_slug'), Tag.name.alias('tag_name'))
            .join(SeriesTag)
            .join(Tag)
            .switch(Series)
            .join(Episode)
            .join(Release)
            .join(subquery, on=(
                subquery.c.id == Series.id))
            .where((Tag.type == TAG_GENRE) & (subquery.c.max_season-Episode.season < MAX_SEASONS))
            .group_by(Tag.name, Series.id)
            .order_by(Tag.name, Series.popularity.desc())
            # Reconstruct objects graph 
            .objects()
            )


def get_series_for_tag(tag):
    subquery = get_series_subquery()
    return (Series.select(Series)
            .join(SeriesTag)
            .switch(Series)
            .join(Episode)
            .join(Release)
            .join(subquery, on=(
                subquery.c.id == Series.id))
            .where((SeriesTag.tag == tag) & (subquery.c.max_season-Episode.season < MAX_SEASONS))
            .order_by(fn.Lower(Series.sort_name))
            .group_by(Series.id)
            )


def get_series_for_language(language):
    subquery = get_series_subquery()
    return (Series.select(Series)
            .switch(Series)
            .join(Episode)
            .join(Release)
            .join(subquery, on=(
                subquery.c.id == Series.id))
            .where((Series.language == language) & (subquery.c.max_season-Episode.season < MAX_SEASONS))
            .order_by(fn.Lower(Series.sort_name))
            .group_by(Series.id)
            )


def release_cte(resolution, size_sorting):
    if size_sorting == "asc":
        aggr_fn = fn.Min(Release.size).alias("size")
    elif size_sorting == "desc":
        aggr_fn= fn.Max(Release.size).alias("size")
    else:
        # Default to sorting by seeders
        aggr_fn = fn.Max(Release.seeders).alias("max_seeders") 

    query = (Release.select(Release.id.alias('release_id'), aggr_fn)
             .join(Episode)
             .where((Release.resolution == resolution) if resolution else True)
             .group_by(Episode.id)
             .cte('best_release'))
    return query


def get_series_with_ids(ids):
    return (Series.select()
            .where(Series.id.in_(ids))
            # .order_by(Series.id)
            )

# ------------
# Search
# ------------

def search_series(query):
    return SeriesIndex.search(f"{query}*", 
                              weights={SeriesIndex.name: 2.0, SeriesIndex.content: 1.0})


def suggest_series(query):
    query = (SeriesIndex
             .select(SeriesIndex, SeriesIndex.bm25().alias('search_score'))
             .where(SeriesIndex.name.match(f"{query}*"))
             .order_by(SeriesIndex.bm25()))
    return query



from pathlib import Path
import pytest
from videobox import create_app
from videobox.models import db_wrapper, Series, Release
from .conftest import TestingConfig

TEST_DIR = Path(__file__).parent

@pytest.fixture()
def app():
    app = create_app(data_dir=TEST_DIR, config_class=TestingConfig)
    with open(TEST_DIR.joinpath("test-data.sql"), 'r') as f:
        # https://docs.python.org/3/library/sqlite3.html#sqlite3.Connection.executescript
        sql = f.read()
        db_wrapper.database.connection().executescript(sql)

    yield app

    db_wrapper.database.connection().close()


@pytest.fixture()
def client(app):
    return app.test_client()

def test_home(client):  
    r = client.get('/')
    assert r.status_code == 200
    # Check main homepage sections
    assert b'Featured series' in r.data 
    assert b'Popular tags' in r.data 
    # Check footer
    assert b'This is Videobox' in r.data 

def test_tag_index(client):  
    r = client.get('/tag')
    assert r.status_code == 200
    # Check main categories
    assert b'Action &amp; Adventure' in r.data 
    assert b'Animation' in r.data 
    assert b'Comedy' in r.data 

def test_tag_detail(client):  
    r = client.get('/tag/action')
    assert r.status_code == 200
    assert b'Action &amp; Adventure' in r.data 

def test_series_detail(client):  
    # Grab a series 
    series = Series.select().get_or_none()    
    r = client.get(f'/series/{series.id}')
    assert r.status_code == 200
    assert series.name in r.text

def test_series_detail_follow(client):
    # Grab a series to follow   
    series = Series.select().get_or_none()
    r = client.post(f'/series/follow/{series.id}', data={'following': 1})
    assert r.status_code == 200
    assert b'Unfollow' in r.data     

def test_search(client):  
    r = client.get('/search', query_string={'query': 'The Simpsons'})
    assert r.status_code == 200
    assert b'The Simpsons' in r.data 

def test_search_info_hash(client):  
    # Grab a release
    release = Release.select().limit(1)[0]
    r = client.get('/search', query_string={'query': release.info_hash})
    assert r.status_code == 302

def test_system_status(client):  
    r = client.get('/status')
    assert r.status_code == 200
    assert b'System status' in r.data


"""
Videobox package.
"""

__version__ = "0.7.1"

import sys
import os
import signal
from pathlib import Path
import click
import flask 
import sqlite3
import waitress
import uuid
import videobox.models as models
import videobox.filters as filters
from .main import bp as main_blueprint
from videobox.main.announcer import announcer
import videobox.sync as sync
import videobox.scraper as scraper
import tomli_w
try:
    import tomllib as toml  # Python 3.11+
except ImportError:
    import tomli as toml

DATABASE_FILENAME = 'library.db'
CONFIG_FILENAME = 'config.toml'
DEFAULT_DATA_DIR = Path.home().joinpath(".videobox")
MAX_WORKER_TIMEOUT = 10 # Seconds

def create_app(app_dir=None, data_dir=None, config_class=None):
    if app_dir:
        app = flask.Flask(__name__, template_folder=os.path.join(app_dir, "templates"), static_folder=os.path.join(app_dir, "static"))
    else:
        app = flask.Flask(__name__)

    if data_dir:
        data_dir = Path(data_dir)
    else:
        data_dir = DEFAULT_DATA_DIR

    os.makedirs(data_dir, exist_ok=True)

    if config_class:    
        app.config.from_object(config_class)
    else:
        app.config['DATABASE_URL'] = f'sqlite:///{data_dir.joinpath(DATABASE_FILENAME)}' 
        config_path = os.path.join(data_dir, CONFIG_FILENAME)
        if app.config.from_file(config_path, load=toml.load, text=False, silent=True):
            app.logger.debug(f"Using configuration file {config_path}")
        else:
            config = get_default_config()
            app.config.from_mapping(config)
            with open(config_path, "wb") as f:
                tomli_w.dump(config, f)

    # Initialize Flask extensions here
    
    models.db_wrapper.init_app(app)
    models.db_wrapper.database.pragma('foreign_keys', 1, permanent=True)
    models.db_wrapper.database.pragma('journal_mode', 'wal', permanent=True)
    # Create db custom functions
    models.db_wrapper.database.register_function(scraper.get_scrape_threshold, "threshold", num_params=1)

    # Make sure db schema is updated but not while testing
    if not app.config['TESTING']:
        migrate_count = models.setup()
        if migrate_count:
            app.logger.debug(f"Changed {migrate_count} database schema fields")

    app.logger.debug(f"Using SQLite {sqlite3.sqlite_version} with database {app.config['DATABASE_URL']}")
    app.logger.debug(f"Client ID is {app.config['API_CLIENT_ID']}")

    # Register main app routes
    app.register_blueprint(main_blueprint)

    # Register custom template filters
    filters.init_app(app)

    with app.app_context():
        def on_update_progress(message):
            data = flask.render_template(
                "_update-progress.html", message=message)            
            msg = announcer.format_sse(data=data, event='sync-progress')
            announcer.announce(msg)

        def on_update_done(message, alert):
            # @@TODO save alert
            data = flask.render_template(
                "_update-done.html", message=message)                    
            msg = announcer.format_sse(data=data, event='sync-done')
            announcer.announce(msg)
            announcer.close()

        # Start immediately
        sync.sync_worker = sync.SyncWorker(app.config["API_CLIENT_ID"], progress_callback=on_update_progress, done_callback=on_update_done)
        # Do not keep syncing while testing
        if not app.config['TESTING']:
            sync.sync_worker.start()

    def handle_shutdown_signal(s, _):
        app.logger.debug(f"Got signal {s}, now stop workers...")
        shutdown_workers(app)
        sys.exit()

    # Install handlers on this thread only if running within the flask dev server/waitress process
    if not app_dir:
        for s in (signal.SIGINT, signal.SIGTERM, signal.SIGQUIT, signal.SIGHUP):
            signal.signal(s, handle_shutdown_signal)

    return app

def shutdown_workers(app):
    # Shutdown all worker threads
    for worker in [sync.sync_worker]:        
        if worker:
            worker.abort()
            if worker.is_alive():
                app.logger.debug(f"Waiting for {worker.name} #{id(worker)} thread to finish work...")
                worker.join(MAX_WORKER_TIMEOUT)


def get_default_config():
    return {"API_CLIENT_ID": uuid.uuid4().hex}

@click.command()
@click.option('--host', help='Hostname or IP address on which to listen, default is 0.0.0.0, which means "all IP addresses on this host".', default="0.0.0.0")
@click.option('--port', help='TCP port on which to listen, default is 8080', type=int, default=8080)
def serve(host, port):
    print(f'Videobox has started. Point your browser to http://{"localhost" if host == "0.0.0.0" else host}:{port} to use the web interface.')
    waitress.serve(create_app(), host=host, port=port, threads=8)



import requests
from flask import current_app
from urllib.parse import urljoin
import videobox

TIMEOUT = 10
API_VERSION = 3
DEFAULT_API_BASE_URL = "https://www.videobox.passiomatic.com/"
USER_AGENT = f"Videobox/{videobox.__version__} (https://pypi.org/project/videobox/)"

# Full import


def get_all_tags(client_id):
    return get_url(f"{API_VERSION}/tags/all?client={client_id}")


def get_all_series(client_id):
    return get_url(f"{API_VERSION}/series/all?client={client_id}")


def get_all_series_tags(client_id):
    return get_url(f"{API_VERSION}/series-tags/all?client={client_id}")


def get_all_episodes(client_id):
    return get_url(f"{API_VERSION}/episodes/all?client={client_id}")


def get_all_releases(client_id):
    return get_url(f"{API_VERSION}/releases/all?client={client_id}")


# Sync


def get_updated_series(client_id, since):
    return get_url(f"{API_VERSION}/series/updated?since={int(since.timestamp())}&client={client_id}")


def get_tags_with_ids(client_id, ids):
    return get_url(f"{API_VERSION}/tags?ids={make_ids(ids)}&client={client_id}")


def get_series_with_ids(client_id, ids):
    return get_url(f"{API_VERSION}/series?ids={make_ids(ids)}&client={client_id}")


def get_series_tags_for_ids(client_id, ids):
    return get_url(f"{API_VERSION}/series-tags?ids={make_ids(ids)}&client={client_id}")


def get_episodes_with_ids(client_id, ids):
    return get_url(f"{API_VERSION}/episodes?ids={make_ids(ids)}&client={client_id}")


def get_releases_with_ids(client_id, ids):
    return get_url(f"{API_VERSION}/releases?ids={make_ids(ids)}&client={client_id}")


def make_ids(ids):
    return ",".join(map(str, ids))


def get_url(path):
    request_headers = {
        'User-Agent': USER_AGENT
    }
    url = urljoin(current_app.config.get('API_BASE_URL', DEFAULT_API_BASE_URL), path)
    current_app.logger.debug(f"Quering API endpoint {url}...")
    return requests.get(url, timeout=TIMEOUT, headers=request_headers)
